import numpy as np
import json, os

from typing import List, Tuple
from scipy.optimize import linprog
from collections import defaultdict, deque
from geopy.distance import geodesic
from shapely import Polygon

from structs.landmarks import Landmark

    
# Function to print the result
def print_res(L: List[Landmark]):

    print('The following order is suggested : ')

    dist = 0
    for elem in L : 
        if elem.time_to_reach_next is not None :
            print('- ' + elem.name + ', time to reach next = ' + str(elem.time_to_reach_next))
            dist += elem.time_to_reach_next
        else : 
            print('- ' + elem.name)

    print("\nMinutes walked : " + str(dist))
    print(f"Visited {len(L)-2} landmarks")


# Prevent the use of a particular solution
def prevent_config(resx):
    
    for i, elem in enumerate(resx):
        resx[i] = round(elem)
    
    N = len(resx)               # Number of edges
    L = int(np.sqrt(N))         # Number of landmarks

    nonzeroind = np.nonzero(resx)[0]                    # the return is a little funky so I use the [0]
    nonzero_tup = np.unravel_index(nonzeroind, (L,L))

    ind_a = nonzero_tup[0].tolist()
    vertices_visited = ind_a
    vertices_visited.remove(0)

    ones = [1]*L
    h = [0]*N
    for i in range(L) :
        if i in vertices_visited :
            h[i*L:i*L+L] = ones

    return h, [len(vertices_visited)-1]


# Prevents the creation of the same circle (both directions)
def prevent_circle(circle_vertices: list, L: int) :

    l1 = [0]*L*L
    l2 = [0]*L*L
    for i, node in enumerate(circle_vertices[:-1]) :
        next = circle_vertices[i+1]

        l1[node*L + next] = 1
        l2[next*L + node] = 1

    s = circle_vertices[0]
    g = circle_vertices[-1]

    l1[g*L + s] = 1
    l2[s*L + g] = 1

    return np.vstack((l1, l2)), [0, 0]


# Returns the order of visit as well as any circles if there are some
def is_connected(resx) :

    # first round the results to have only 0-1 values
    for i, elem in enumerate(resx):
        resx[i] = round(elem)
    
    N = len(resx)               # length of res
    L = int(np.sqrt(N))         # number of landmarks. CAST INTO INT but should not be a problem because N = L**2 by def.
    n_edges = resx.sum()        # number of edges

    nonzeroind = np.nonzero(resx)[0] # the return is a little funny so I use the [0]

    nonzero_tup = np.unravel_index(nonzeroind, (L,L))

    ind_a = nonzero_tup[0].tolist()
    ind_b = nonzero_tup[1].tolist()

    # Step 1: Create a graph representation
    graph = defaultdict(list)
    for a, b in zip(ind_a, ind_b):
        graph[a].append(b)

    # Step 2: Function to perform BFS/DFS to extract journeys
    def get_journey(start):
        journey_nodes = []
        visited = set()
        stack = deque([start])

        while stack:
            node = stack.pop()
            if node not in visited:
                visited.add(node)
                journey_nodes.append(node)
                for neighbor in graph[node]:
                    if neighbor not in visited:
                        stack.append(neighbor)

        return journey_nodes

    # Step 3: Extract all journeys
    all_journeys_nodes = []
    visited_nodes = set()

    for node in ind_a:
        if node not in visited_nodes:
            journey_nodes = get_journey(node)
            all_journeys_nodes.append(journey_nodes)
            visited_nodes.update(journey_nodes)

    for l in all_journeys_nodes :
        if 0 in l :
            order = l
            all_journeys_nodes.remove(l)
            break

    if len(all_journeys_nodes) == 0 :
        return order, None

    return order, all_journeys_nodes


# Function that returns the time in minutes from one location to another
def get_time(p1: Tuple[float, float], p2: Tuple[float, float], detour: float, speed: float) :
    
    # Compute the straight-line distance in km
    if p1 == p2 :
        return 0
    else: 
        dist = geodesic(p1, p2).kilometers

    # Consider the detour factor for average cityto deterline walking distance (in km)
    walk_dist = dist*detour

    # Time to walk this distance (in minutes)
    walk_time = walk_dist/speed*60

    return round(walk_time)


# Initialize A and c. Compute the distances from all landmarks to each other and store attractiveness
# We want to maximize the sightseeing :  max(c) st. A*x < b   and   A_eq*x = b_eq
def init_ub_dist(landmarks: List[Landmark], max_steps: int):
    
    with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
        parameters = json.loads(f.read())
        detour = parameters['detour factor']
        speed = parameters['average walking speed']
    
    # Objective function coefficients. a*x1 + b*x2 + c*x3 + ...
    c = []
    # Coefficients of inequality constraints (left-hand side)
    A_ub = []

    for spot1 in landmarks :
        dist_table = [0]*len(landmarks)
        c.append(-spot1.attractiveness)
        for j, spot2 in enumerate(landmarks) :
            t = get_time(spot1.location, spot2.location, detour, speed)
            dist_table[j] = t
        closest = sorted(dist_table)[:22]
        for i, dist in enumerate(dist_table) :
            if dist not in closest :
                dist_table[i] = 32700
        A_ub += dist_table
    c = c*len(landmarks)

    return c, A_ub, [max_steps]


# Constraint to respect only one travel per landmark. Also caps the total number of visited landmarks
def respect_number(L: int, max_landmarks: int):

    ones = [1]*L
    zeros = [0]*L
    A = ones + zeros*(L-1)
    b = [1]
    for i in range(L-1) :
        h_new = zeros*i + ones + zeros*(L-1-i)
        A = np.vstack((A, h_new))
        b.append(1)

    if max_landmarks is None :
        # Read the parameters from the file
        with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
            parameters = json.loads(f.read())
            max_landmarks = parameters['max landmarks']

    A = np.vstack((A, ones*L))
    b.append(max_landmarks+1)

    return A, b


# Constraint to not have d14 and d41 simultaneously. Does not prevent cyclic paths with more elements
def break_sym(L):

    upper_ind = np.triu_indices(L,0,L)

    up_ind_x = upper_ind[0]
    up_ind_y = upper_ind[1]

    A = [0]*L*L         # useless row to prevent overhead ? better solution welcomed
    # A[up_ind_x[0]*L + up_ind_y[0]] = 1
    # A[up_ind_y[0]*L + up_ind_x[0]] = 1
    b = [1]

    for i, _ in enumerate(up_ind_x[1:]) :
        l = [0]*L*L
        if up_ind_x[i] != up_ind_y[i] :
            l[up_ind_x[i]*L + up_ind_y[i]] = 1
            l[up_ind_y[i]*L + up_ind_x[i]] = 1

            A = np.vstack((A,l))
            b.append(1)

    return A, b


# Constraint to not stay in position. Removes d11, d22, d33, etc.
def init_eq_not_stay(L: int): 
    l = [0]*L*L

    for i in range(L) :
        for j in range(L) :
            if j == i :
                l[j + i*L] = 1
    
    l = np.array(np.array(l), dtype=np.int8)

    return [l], [0]


# Go through the landmarks and force the optimizer to use landmarks marked as must_do
def respect_user_must_do(landmarks: List[Landmark]) :
    L = len(landmarks)

    A = [0]*L*L
    b = [0]
    for i, elem in enumerate(landmarks[1:]) :
        if elem.must_do is True and elem.name not in ['finish', 'start']:
            l = [0]*L*L
            l[i*L:i*L+L] = [1]*L        # set mandatory departures from landmarks tagged as 'must_do'

            A = np.vstack((A,l))
            b.append(1)

    return A, b


# Constraint to ensure start at start and finish at goal
def respect_start_finish(L: int):
    l_start = [1]*L + [0]*L*(L-1)   # sets departures only for start (horizontal ones)
    l_start[L-1] = 0                # prevents the jump from start to finish
    l_goal = [0]*L*L                # sets arrivals only for finish (vertical ones)
    l_L = [0]*L*(L-1) + [1]*L       # prevents arrivals at start and departures from goal
    for k in range(L-1) :           # sets only vertical ones for goal (go to)
        l_L[k*L] = 1
        if k != 0 :
            l_goal[k*L+L-1] = 1 
            

    A = np.vstack((l_start, l_goal))
    b = [1, 1]
    A = np.vstack((A,l_L))
    b.append(0)

    return A, b


# Constraint to tie the problem together. Necessary but not sufficient to avoid circles
def respect_order(L: int): 

    A = [0]*L*L                     # useless row to reduce overhead ? better solution is welcome
    b = [0]
    for i in range(L-1) :           # Prevent stacked ones
        if i == 0 or i == L-1:      # Don't touch start or finish
            continue
        else : 
            l = [0]*L
            l[i] = -1
            l = l*L
            for j in range(L) :
                l[i*L + j] = 1

            A = np.vstack((A,l))
            b.append(0)

    return A, b


# Computes the time to reach from each landmark to the next
def link_list(order: List[int], landmarks: List[Landmark])->List[Landmark] :
    
    # Read the parameters from the file
    with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
        parameters = json.loads(f.read())
        detour_factor = parameters['detour factor']
        speed = parameters['average walking speed']
    
    L =  []
    j = 0
    total_dist = 0
    while j < len(order)-1 :
        elem = landmarks[order[j]]
        next = landmarks[order[j+1]]

        d = get_time(elem.location, next.location, detour_factor, speed)
        elem.time_to_reach_next = d
        elem.must_do = True
        elem.location = (round(elem.location[0], 5), round(elem.location[1], 5))
        L.append(elem)
        j += 1
        total_dist += d

    next.location = (round(next.location[0], 5), round(next.location[1], 5))   
    L.append(next)
    
    return L, total_dist


# Same as link_list but does it on a already ordered list
def link_list_simple(ordered_visit: List[Landmark])-> List[Landmark] :
    
    # Read the parameters from the file
    with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
        parameters = json.loads(f.read())
        detour_factor = parameters['detour factor']
        speed = parameters['average walking speed']

    L =  []
    j = 0
    total_dist = 0
    while j < len(ordered_visit)-1 :
        elem = ordered_visit[j]
        next = ordered_visit[j+1]

        elem.next_uuid = next.uuid
        d = get_time(elem.location, next.location, detour_factor, speed)
        elem.time_to_reach_next = d
        if elem.name not in ['start', 'finish'] :
            elem.must_do = True
        L.append(elem)
        j += 1
        total_dist += d

    L.append(next)
    
    return L, total_dist


# Main optimization pipeline
def solve_optimization (landmarks :List[Landmark], max_steps: int, printing_details: bool, max_landmarks = None) :

    L = len(landmarks)

    # SET CONSTRAINTS FOR INEQUALITY
    c, A_ub, b_ub = init_ub_dist(landmarks, max_steps)          # Add the distances from each landmark to the other
    A, b = respect_number(L, max_landmarks)                     # Respect max number of visits (no more possible stops than landmarks). 
    A_ub = np.vstack((A_ub, A), dtype=np.int16)
    b_ub += b
    A, b = break_sym(L)                                         # break the 'zig-zag' symmetry
    A_ub = np.vstack((A_ub, A), dtype=np.int16)
    b_ub += b


    # SET CONSTRAINTS FOR EQUALITY
    A_eq, b_eq = init_eq_not_stay(L)                            # Force solution not to stay in same place
    A, b = respect_user_must_do(landmarks)                      # Check if there are user_defined must_see. Also takes care of start/goal
    A_eq = np.vstack((A_eq, A), dtype=np.int8)
    b_eq += b
    A, b = respect_start_finish(L)                  # Force start and finish positions
    A_eq = np.vstack((A_eq, A), dtype=np.int8)
    b_eq += b
    A, b = respect_order(L)                         # Respect order of visit (only works when max_steps is limiting factor)
    A_eq = np.vstack((A_eq, A), dtype=np.int8)
    b_eq += b
    
    # SET BOUNDS FOR DECISION VARIABLE (x can only be 0 or 1)
    x_bounds = [(0, 1)]*L*L

    # Solve linear programming problem
    res = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq = b_eq, bounds=x_bounds, method='highs', integrality=3)

    # Raise error if no solution is found
    if not res.success :
            raise ArithmeticError("No solution could be found, the problem is overconstrained. Please adapt your must_dos")

    # If there is a solution, we're good to go, just check for connectiveness
    else :
        order, circles = is_connected(res.x)
        #nodes, edges = is_connected(res.x)
        i = 0
        timeout = 80
        while circles is not None and i < timeout:
            A, b = prevent_config(res.x)
            A_ub = np.vstack((A_ub, A))
            b_ub += b
            #A_ub, b_ub = prevent_circle(order, len(landmarks), A_ub, b_ub)
            for circle in circles :
                A, b = prevent_circle(circle, len(landmarks))
                A_eq = np.vstack((A_eq, A))
                b_eq += b
            res = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq = b_eq, bounds=x_bounds, method='highs', integrality=3)
            if not res.success :
                print("Solving failed because of overconstrained problem")
                return None
            order, circles = is_connected(res.x)
            #nodes, edges = is_connected(res.x)
            if circles is None :
                break
            print(i)
            i += 1
        
        if i == timeout :
            raise TimeoutError(f"Optimization took too long. No solution found after {timeout} iterations.")

        # Add the times to reach and stop optimizing
        L, _ = link_list(order, landmarks)
            
        if printing_details is True :
            if i != 0 :
                print(f"Neded to recompute paths {i} times because of unconnected loops...")
            print_res(L)
            print("\nTotal score : " + str(int(-res.fun)))

        return L






